'use strict';

var react = require('react');
var proxyCompare = require('proxy-compare');
var vanilla = require('valtio/vanilla');

var useAffectedDebugValue = function useAffectedDebugValue(state, affected) {
  var pathList = react.useRef();
  react.useEffect(function () {
    pathList.current = proxyCompare.affectedToPathList(state, affected, true);
  });
  react.useDebugValue(pathList.current);
};
var targetCache = new WeakMap();
function useSnapshot(proxyObject, options) {
  var notifyInSync = options == null ? void 0 : options.sync;
  var lastSnapshot = react.useRef();
  var lastAffected = react.useRef();
  var inRender = true;
  var currSnapshot = react.useSyncExternalStore(react.useCallback(function (callback) {
    var unsub = vanilla.subscribe(proxyObject, callback, notifyInSync);
    callback();
    return unsub;
  }, [proxyObject, notifyInSync]), function () {
    var nextSnapshot = vanilla.snapshot(proxyObject);
    try {
      if (!inRender && lastSnapshot.current && lastAffected.current && !proxyCompare.isChanged(lastSnapshot.current, nextSnapshot, lastAffected.current, new WeakMap())) {
        return lastSnapshot.current;
      }
    } catch (e) {}
    return nextSnapshot;
  }, function () {
    return vanilla.snapshot(proxyObject);
  });
  inRender = false;
  var currAffected = new WeakMap();
  react.useEffect(function () {
    lastSnapshot.current = currSnapshot;
    lastAffected.current = currAffected;
  });
  if (process.env.NODE_ENV !== 'production') {
    useAffectedDebugValue(currSnapshot, currAffected);
  }
  var proxyCache = react.useMemo(function () {
    return new WeakMap();
  }, []);
  return proxyCompare.createProxy(currSnapshot, currAffected, proxyCache, targetCache);
}

exports.useSnapshot = useSnapshot;
